// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import BigInt from 'big-integer';
import Utils from './utils';
import C25519Point from './c25519Point';
import C25519Key from './c25519Key';
import Hash from './hash';
import SecretShare from './secretShare';
import { getBig } from './bnInput';

export default class C25519Cipher {
    /**
     * @param {C25519Point} c1
     * @param {Uint8Array} c2
     */
    constructor(c1, c2) {
        this.c1 = c1;
        this.c2 = c2;
    }

    toString() {
        return Buffer.from(this.toArray()).toString("base64");
    }

    toArray() {
        return Buffer.concat([Buffer.from(this.c2), Buffer.from(this.c1.toArray())]);
    }

    /** @param {C25519Key} key */
    decrypt(key) {
        return xor(Hash.shaBuffer(this.c1.times(key.x).toArray()), this.c2);
    }
    
    /** @param {C25519Key} key */
    partialDecrypt(key) {
        return new C25519Cipher(this.c1.times(key.x), this.c2);
    }

    /**
     * @param {C25519Cipher[]} cyphers
     * @param {import('./bnInput').BNInput[]} ids
     */
    static decryptShares(cyphers, ids)
    {
        var nums = ids.map(id => getBig(id));
        var c1s = nums.map(id => SecretShare.getLi(id, nums, C25519Point.n))
            .map((li, i) => cyphers[i].c1.times(li));

        var c1 = c1s.reduce((sum, partial) => sum.add(partial), C25519Point.infinity);

        return xor(Hash.shaBuffer(c1.toArray()), cyphers[0].c2);
    }

    /**
     * @param {Uint8Array} m
     * @param {C25519Key} key
     */
    static encrypt(m, key) {
        const k = Utils.random(BigInt.one, C25519Point.n.subtract(BigInt.one));
        const c1 = C25519Point.g.times(k);
        const c2 = xor(Hash.shaBuffer(key.y.times(k).toArray()), m);

        return new C25519Cipher(c1, c2);
    }

/** @param {Uint8Array} data */
    static from(data) {
        const c2 = data.slice(0, 32);
        const c1 = C25519Point.from(data.slice(32));
        return new C25519Cipher(c1, c2);
    }

    /** @param {string} message */
    static fromString(message) {
        return C25519Cipher.from(Buffer.from(message, "base64"))
    }
}

/**
 * @param {Uint8Array} data1
 * @param {Uint8Array} data2
 */
function xor(data1, data2)
{
    const result = Buffer.alloc(data1.length);
    for (let i = 0; i < data1.length; i++)
        result[i] = (data1[i] ^ data2[i]);

    return result;
}
