// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import BigInt from 'big-integer';
import Utils from './utils';
import C25519Point from './c25519Point';
import { getBig, getArray } from './bnInput';
import C25519Cipher from './c25519Cipher';
import AesSherableKey from './aesSherableKey';
import SecretShare from './secretShare';

export default class C25519Key {
    /**
     * @param {import('./bnInput').BNInput} x
     * @param {C25519Point} y
     */
    constructor(x, y) {
        this.x = getBig(x);
        this.y = y;
    }
    
    public() {
        return new C25519Key(0, this.y);
    }

    /** @param {AesSherableKey} key */
    encryptKey(key) {
        return this.encrypt(key.toArray());
    }
    
    /** @param {Uint8Array} m */
    encrypt(m) {
        return C25519Cipher.encrypt(m, this);
    }
    
    /** @param {Uint8Array|string} cipher */
    decryptKey(cipher) {
        return AesSherableKey.from(this.decrypt(cipher));
    }
    
    /** @param {Uint8Array|string} data */
    decrypt(data) {
        const cipher = fromBase64(data);
        return C25519Cipher.from(cipher).decrypt(this);
    }
    
    /** @param {Uint8Array|string} data */
    partialDecrypt(data) {
        const cipher = fromBase64(data);
        return C25519Cipher.from(cipher).partialDecrypt(this);
    }

    /**
     * @param {number} threshold 
     * @param {import('./bnInput').BNInput[]} ids
     * @param {boolean} noPublic 
     */
    share(threshold, ids, noPublic = false) {
        const nums = ids.map(id => getBig(id));
        var [,shrs] = SecretShare.shareFromIds(this.x, nums, threshold, C25519Point.n);
        return shrs.map(shr => C25519Key.private(shr, noPublic));
    }

    toString() {
        return Buffer.from(this.toArray()).toString("base64");
    }

    toArray() {
        const x = Buffer.from(Utils.padLeft(getArray(this.x), 32));
        return Buffer.concat([x, Buffer.from(this.y.toArray())]);
    }

    /**
     * @param {import('./bnInput').BNInput} x
     * @param {boolean} noPublic
     */
    static private(x, noPublic = false) {
        return new C25519Key(x, noPublic ? C25519Point.infinity : C25519Point.g.times(x));
    }

    static generate() {
        const x = Utils.random(BigInt.one, C25519Point.n.subtract(BigInt.one));
        return new C25519Key(x, C25519Point.g.times(x));
    }

    /** @param {string | Uint8Array} data */
    static from(data) {
        const buffer = typeof data === "string"
            ? Buffer.from(data, "base64") : Buffer.from(data);

        const x = getBig(buffer.slice(0, 32));
        const y = C25519Point.from(buffer.slice(32));
        return new C25519Key(x, y);
    }

    /** @param {string} message */
    static fromString(message) {
        return C25519Key.from(Buffer.from(message, "base64"))
    }
}

/** 
 * @param {string|Uint8Array} data
 * @return {Uint8Array}
 */
function fromBase64(data) {
    return typeof data === "string" ? Buffer.from(data, "base64") : data;
}