// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import assert from 'assert';
import utils, { constants } from './utils.js';
import secretShare from './secretShare.js';
import ElGamalKey from './elGamalKey.js';

export default class ElGamal {
    static encrypt(data, pub) {
        return ElGamal.encryptBuffer(data, pub).toString('base64')
    }

    static encryptBuffer(data, pub) {
        const chunks = utils.decode(data, pub.size, 'utf8');
        let numbers = [];

        for (let i = 0; i < chunks.length; i++) {
            const r = utils.random(constants.zero, pub.p.subtract(constants.one));
            numbers.push(pub.g.modPow(r, pub.p)); //C1
            numbers.push(pub.key.modPow(r, pub.p).multiply(chunks[i]).mod(pub.p)); //C2
        }

        return Buffer.from(utils.encodeByteArray(numbers, pub.size, true));
    }

    /**
     * 
     * @param {string} cypher 
     * @param {ElGamalKey} prv
     */
    static decrypt(cypher, prv) {
        try {
            return ElGamal.decryptBuffer(cypher, prv).toString('utf8');
        } catch { return ""; }
    }

    static decryptBuffer(cypher, prv) {
        const bytes = !utils.isBuffer(cypher) ? Buffer.from(cypher, 'base64') : cypher;
        assert(bytes.length % (prv.size / 8 * 2) === 0, 'Malformed Cipher Text');

        const chunks = utils.decode(bytes, prv.size);
        let numbers = [];
        for (let i = 0; i < chunks.length; i += 2) {
            const c1 = chunks[i];
            const c2 = chunks[i + 1];
            numbers.push(c1.modPow(prv.key, prv.p).modInv(prv.p).multiply(c2).mod(prv.p));
        }

        return Buffer.from(utils.encodeByteArray(numbers, prv.size));
    }

    static decryptPartial(cypher, prv) {
        return ElGamal.decryptPartialBuffer(cypher, prv).toString('base64');
    }

    static decryptPartialBuffer(cypher, prv) {
        const bytes = !utils.isBuffer(cypher) ? Buffer.from(cypher, 'base64') : cypher;
        assert(bytes.length % (prv.size / 8 * 2) === 0, 'Malformed Cipher Text');

        const chunks = utils.decode(bytes, prv.size);
        let numbers = [prv.id];
        for (let i = 0; i < chunks.length; i += 2) {
            numbers.push(chunks[i].modPow(prv.key, prv.p)); //C1 ^ X'
            numbers.push(chunks[i + 1]); //C2
        }

        return Buffer.from(utils.encodeByteArray(numbers, prv.size, true));
    }
    
    static combinePartial(partials, pub) {
        try {
            return ElGamal.combinePartialBuffer(partials, pub).toString('utf8');
        } catch { return ""; }
    }

    static combinePartialBuffer(partials, pub) {
        //TODO: Data validation
        const shares = partials.map(p => utils.decode(p, pub.size, 'base64'));
        const ids = shares.map(sh => sh[0]);
        const first = shares[0]
        const numbers = [];
        for (let i = 1; i < first.length; i += 2) {
            const ss = shares.map(sh => sh[i]);
            const s = secretShare.interpolatePartial(ids, ss, pub.p, pub.q);
            const c2 = first[i + 1];
            numbers.push(s.modInv(pub.p).multiply(c2).mod(pub.p));
        }

        return Buffer.from(utils.encodeByteArray(numbers, pub.size));
    }
}
